# __init__.py
#
# Folder 'dictionary'
#
# Copyright (c) 2022-2024 WordTM Project 
# Author: Johnny Cheng <drjohnnycheng@gmail.com>
#
# Updated: 2 February 2024
#
# URL: https://github.com/drjohnnycheng/wordtm.git
# For license information, see LICENSE.TXT

"""The folder stores the dictionaries of Chinese (Traditional) stopwords
and Chinese terms for Jieba word segmentation.
"""
