# version.py
#    
# Version information of the module WordTM
#
# Copyright (c) 2022-2024 - Johnny Cheng <drjohnnycheng@gmail.com>
#
# Updated: 2 February 2024
#
# URL: https://github.com/drjohnnycheng/wordtm.git
# For license information, see LICENSE.TXT

__author__ = "Johnny Cheng"
__copyright__ = "Copyright (c) 2022-2024 - Johnny Cheng <drjohnnycheng@gmail.com>"
__credits__ = ["Jehovah, the Lord"]
__license__ = ""
__version__ = "0.4.3"
__email__ = "drjohnnycheng@gmail.com"
__status__ = "Development"
__url__ = "https://github.com/drjohnnycheng/wordtm.git"
